# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['negmas',
 'negmas.apps',
 'negmas.apps.scml',
 'negmas.external',
 'negmas.scripts',
 'negmas.tests']

package_data = \
{'': ['*'],
 'negmas': ['logs/*'],
 'negmas.tests': ['config/*',
                  'data/10issues/*',
                  'data/AMPOvsCity/*',
                  'data/Laptop/*',
                  'data/Laptop1Issue/*',
                  'data/LaptopConv/*',
                  'data/LaptopConv1D/*',
                  'data/fuzzyagent/*',
                  'data/scenarios/anac/y2010/EnglandZimbabwe/*',
                  'data/scenarios/anac/y2010/ItexvsCypress/*',
                  'data/scenarios/anac/y2010/Travel/*',
                  'data/scenarios/other/S-1NIKFRT-1/*',
                  'scml/*']}

install_requires = \
['Click>=6.0',
 'PyYAML==5.1b1',
 'colorlog',
 'inflect',
 'matplotlib',
 'numpy>=1.16',
 'pandas>=0.24.1',
 'progressbar2>=3.39',
 'py4j',
 'pytest-runner>=4.4',
 'scipy>=1.2',
 'setuptools>=40.8.0',
 'stringcase',
 'tabulate',
 'tqdm>=4.31.1',
 'typing',
 'typing_extensions>=3.7']

extras_require = \
{':python_version < "3.7"': ['dataclasses']}

entry_points = \
{'console_scripts': ['negmas = negmas.scripts.app:cli']}

setup_kwargs = {
    'name': 'negmas',
    'version': '0.2.5',
    'description': 'NEGotiations Managed by Agent Simulations',
    'long_description': ".. image:: https://img.shields.io/pypi/pyversions/negmas.svg\n        :target: https://pypi.python.org/pypi/negmas\n        :alt: Python\n\n.. image:: https://img.shields.io/pypi/status/negmas.svg\n        :target: https://pypi.python.org/pypi/negmas\n        :alt: Pypi\n\n.. image:: https://img.shields.io/pypi/l/negmas.svg\n        :target: https://pypi.python.org/pypi/negmas\n        :alt: License\n\n.. image:: https://img.shields.io/pypi/dm/negmas.svg\n        :target: https://pypi.python.org/pypi/negmas\n        :alt: Downloads\n\n.. image:: https://img.shields.io/codacy/coverage/1b204fe0a69e41a298a175ea225d7b81.svg\n        :target: https://app.codacy.com/project/yasserfarouk/negmas/dashboard\n        :alt: Coveage\n\n.. image:: https://img.shields.io/codacy/grade/1b204fe0a69e41a298a175ea225d7b81.svg\n        :target: https://app.codacy.com/project/yasserfarouk/negmas/dashboard\n        :alt: Code Quality\n\n.. image:: https://img.shields.io/pypi/v/negmas.svg\n        :target: https://pypi.python.org/pypi/negmas\n        :alt: Pypi\n\n.. image:: https://img.shields.io/travis/yasserfarouk/negmas.svg\n        :target: https://travis-ci.org/yasserfarouk/negmas\n        :alt: Build Status\n\n.. image:: https://readthedocs.org/projects/negmas/badge/?version=latest\n        :target: https://negmas/readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n\n\nNegMAS is a python library for developing autonomous negotiation agents embedded in simulation environments.\nThe name ``negmas`` stands for either NEGotiation MultiAgent System or NEGotiations Managed by Agent Simulations\n(your pick). The main goald of NegMAS is to advance the state of the art in situated simultaneous negotiations.\nNevertheless, it can; and was used; in modeling simpler bilateral and multi-lateral negotiations, preference elicitation\n, etc.\n\nIntroduction\n============\n\nThis package was designed to help advance the state-of-art in negotiation research by providing an easy-to-use yet\npowerful platform for autonomous negotiation targeting situated simultaneous negotiations.\nIt grew out of the NEC-AIST collaborative laboratory project.\n\nBy *situated* negotiations, we mean those for which utility functions are not pre-ordained by fiat but are a natural\nresult of a simulated business-like process.\n\nBy *simultaneous* negotiations, we mean sessions of dependent negotiations for which the utility value of an agreement\nof one session is affected by what happens in other sessions.\n\nThe documentation is available at: documentation_\n\n.. _documentation: https://negmas.readthedocs.io/\n\nMain Features\n=============\n\nThis platform was designed with both flexibility and scalability in mind. The key features of the NegMAS package are:\n\n#. The public API is decoupled from internal details allowing for scalable implementations of the same interaction\n   protocols.\n#. Supports agents engaging in multiple concurrent negotiations.\n#. Provides support for inter-negotiation synchronization either through coupled utility functions or through central\n   *control* agents.\n#. The package provides sample negotiators that can be used as templates for more complex negotiators.\n#. The package supports both mediated and unmediated negotiations.\n#. Supports both bilateral and multilateral negotiations.\n#. Novel negotiation protocols and simulated *worlds* can be added to the package as easily as adding novel negotiators.\n#. Allows for non-traditional negotiation scenarios including dynamic entry/exit from the negotiation.\n#. A large variety of built in utility functions.\n#. Utility functions can be active dynamic entities which allows the system to model a much wider range of dynamic ufuns\n   compared with existing packages.\n#. A distributed system with the same interface and industrial-strength implementation is being created allowing agents\n   developed for NegMAS to be seemingly employed in real-world business operations.\n\nTo use negmas in a project\n\n.. code-block:: python\n\n    import negmas\n\nThe package was designed for many uses cases. On one extreme, it can be used by an end user who is interested in running\none of the built-in negotiation protocols. On the other extreme, it can be used to develop novel kinds of negotiation\nagents, negotiation protocols, multi-agent simulations (usually involving situated negotiations), etc.\n\nRunning existing negotiators/negotiation protocols\n==================================================\n\nUsing the package for negotiation can be as simple as the following code snippet:\n\n.. code-block:: python\n\n    from negmas import SAOMechanism, AspirationNegotiator, MappingUtilityFunction\n    session = SAOMechanism(outcomes=10, n_steps=100)\n    negotiators = [AspirationNegotiator(name=f'a{_}') for _ in range(5)]\n    for negotiator in negotiators:\n        session.add(negotiator, ufun=MappingUtilityFunction(lambda x: random.random() * x[0]))\n\n    session.run()\n\nIn this snippet, we created a mechanism session with an outcome-space of *10* discrete outcomes that would run for *10*\nsteps. Five agents with random utility functions are then created and *added* to the session. Finally the session is\n*run* to completion. The agreement (if any) can then be accessed through the *state* member of the session. The library\nprovides several analytic and visualization tools to inspect negotiations. See the first tutorial on\n*Running a Negotiation* for more details.\n\nDeveloping a negotiator\n=======================\n\nDeveloping a novel negotiator slightly more difficult by is still doable in few lines of code:\n\n.. code-block:: python\n\n    from negmas.sao import SAONegotiator\n    from negmas import ResponseType\n    class MyAwsomeNegotiator(SAONegotiator):\n        def __init__(self):\n            # initialize the parents\n            super().__init__(self)\n\n        def respond(self, offer, state):\n            # decide what to do when receiving an offer\n            return ResponseType.ACCEPT_OFFER\n\n        def propose(self, state):\n            # proposed the required number of proposals (or less) \n            pass\n\nBy just implementing `respond()` and `propose()`. This negotiator is now capable of engaging in alternating offers\nnegotiations. See the documentation of `Negotiator` for a full description of available functionality out of the box.\n\nDeveloping a negotiation protocol\n=================================\n\nDeveloping a novel negotiation protocol is actually even simpler:\n\n.. code-block:: python\n\n    from negmas.mechanisms import Mechanism\n\n    class MyNovelProtocol(Mechanism):\n        def __init__(self):\n            super().__init__()\n\n        def round(self):\n            # one step of the protocol\n            pass\n\nBy implementing the single `round()` function, a new protocol is created. New negotiators can be added to the\nnegotiation using `add()` and removed using `remove()`. See the documentation for a full description of\n`Mechanism` available functionality out of the box [Alternatively you can use `Protocol` instead of `Mechanism`].\n\n\nRunning a world simulation\n==========================\n\nThe *raison d'être* for NegMAS is to allow you to develop negotiation agents capable of behaving in realistic\n*business like* simulated environments. These simulations are called *worlds* in NegMAS. Agents interact with each other\nwithin these simulated environments trying to maximize some intrinsic utility function of the agent through several\n*possibly simultaneous* negotiations.\n\nThe `situated` module provides all that you need to create such worlds. An example can be found in the `scml` package.\nThis package implements a supply chain management system in which factory managers compete to maximize their profits in\na market with only negotiations as the means of securing contracts.\n\n\nAcknowledgement\n===============\n\n.. _Genius: http://ii.tudelft.nl/genius\n\nNegMAS tests use scenarios used in ANAC 2010 to ANAC 2018 competitions obtained from the Genius_ Platform. These domains\ncan be found in the tests/data and notebooks/data folders.\n",
    'author': 'Yasser Mohammad',
    'author_email': 'yasserfarouk@gmail.com',
    'url': 'https://github.com/yasserfarouk/negmas',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



import pytest
import random

from .fixtures import *
from negmas import Issues, outcome_is_valid, outcome_in_range


def test_dict_outcomes(issues, valid_outcome_dict, invalid_outcome_dict):
    assert outcome_is_valid(valid_outcome_dict, issues)
    assert not outcome_is_valid(invalid_outcome_dict, issues)


def test_list_outcomes(int_issues, valid_outcome_list, invalid_outcome_list):
    assert outcome_is_valid(valid_outcome_list, int_issues)
    assert not outcome_is_valid(invalid_outcome_list, int_issues)

def test_outcome_in_verious_ranges():
    or1 = {'price': (0.0, 2.0)
        , 'distance': [0.3, 0.4]
        , 'type': ['a', 'b']
        , 'area': 3}
    or2 = {'price': [(0.0, 1.0), (1.5, 2.0)]
        , 'area': [(3, 4), (7, 9)]}

    assert outcome_in_range({'date': '2018.10.4'}, or1)
    assert not outcome_in_range({'date': '2018.10.4'}, or1, strict=True)
    assert outcome_in_range({'area': 3}, or1)
    assert not outcome_in_range({'type': 'c'}, or1)
    assert outcome_in_range({'type': 'a'}, or1)
    assert not outcome_in_range({'date': '2018.10.4'}, or2, strict=True)
    assert not outcome_in_range({'area': 3}, or2)
    assert outcome_in_range({'area': 3.0001}, or2)
    assert not outcome_in_range({'area': 5}, or2)
    assert outcome_in_range({'price': 0.4}, or2)
    assert not outcome_in_range({'price': 1.2}, or2)
    assert not outcome_in_range({'price': 0.4, 'area': 4}, or2)
    assert not outcome_in_range({'price': 0.4, 'area': 10}, or2)
    assert not outcome_in_range({'price': 1.2, 'area': 10}, or2)
    assert not outcome_in_range({'price': 1.2, 'area': 4}, or2)
    assert outcome_in_range({'type': 'a'}, or2)
    or1 = {'price': 10}
    assert outcome_in_range({'price': 10}, or1)
    assert not outcome_in_range({'price': 11}, or1)


def test_issues_construction():
    issues = Issues(price=(0.0, 1.0), cost=[1,2,3], delivery=['yes', 'no'])
    assert len(issues.issues) == 3
    assert str(issues) == "price: (0.0, 1.0)\ncost: [1, 2, 3]\ndelivery: ['yes', 'no']"
    assert issues.is_infinite()
    assert not issues.is_finite()
    assert all(a==b for a, b in zip(issues.types, ['continuous', 'discrete', 'discrete']))

    issues = Issues.from_single_issue(Issue(10, 'issue'))
    assert len(issues.issues) == 1
    assert str(issues) == 'issue: 10'

    issues = Issues(price=[2, 3], cost=[1, 2, 3], delivery=['yes', 'no'])
    assert issues.is_finite()
    assert not issues.is_infinite()
    assert issues.cardinality() == issues.n_outcomes() == 2*3*2

    valid = issues.rand_valid()
    invalid = issues.rand_invalid()
    assert outcome_in_range(valid, issues.outcome_range)
    assert outcome_in_range(invalid, issues.outcome_range)


def test_outcome_constraint():
    r1 = {'price': [(0.0, 1.0), (3.0, 4.0)]
          , 'quantity' : [1, 2, 4, 5]
          , 'delivery': 'yes'
          }

    notr1 = {'price': [(3.01, 3.99), (4.01, 10.0)]
        , 'quantity': 3
        , 'delivery': 'no'
          }

    r2 = {'price': (3.01, 10.0)
        , 'quantity': [2, 3]
        , 'delivery': ['yes', 'no']
          }

    # r = OutcomeSpace(r1)

if __name__ == '__main__':
    pytest.main(args=[__file__])



import functools
import itertools
import math
from abc import ABC
from random import random, randint
from typing import TYPE_CHECKING

from dataclasses import dataclass
from numpy.random import dirichlet

from negmas import AgentMechanismInterface, MechanismState
from negmas.apps.scml import FinancialReport
from negmas.events import Notification
from negmas.helpers import ConfigReader, get_class
from negmas.negotiators import Negotiator
from negmas.outcomes import Issue
from negmas.sao import AspirationNegotiator
from negmas.situated import Contract, Breach
from negmas.situated import RenegotiationRequest
from negmas.utilities import normalize, ComplexWeightedUtilityFunction, MappingUtilityFunction
from .common import CFP
from .agent import SCMLAgent
from .helpers import pos_gauss

if True: #
    from typing import Dict, Any, List, Optional, Union, Tuple
    from .common import Loan

if TYPE_CHECKING:
    from .awi import SCMLAWI

__all__ = [
    'Consumer',
    'ConsumptionProfile',
    'ScheduleDrivenConsumer'
]


@dataclass
class ConsumptionProfile:
    schedule: Union[int, List[int]] = 0
    underconsumption: float = 0.1
    overconsumption: float = 0.01
    dynamicity: float = 0.0
    cv: float = 0.1

    alpha_q: float = 0.5
    alpha_u: float = 1.0

    beta_q: float = 10.0
    beta_u: float = 10.0

    tau_q: float = 2
    tau_u: float = 0.25

    @classmethod
    def random(cls):
        return ConsumptionProfile(schedule=randint(0, 5), overconsumption=2 * random(), underconsumption=2 * random()
                                  , cv=random(), beta_q=99 * random() + 1, beta_u=99 * random() + 1, tau_q=3 * random()
                                  , tau_u=random())

    def schedule_at(self, time: int) -> int:
        if isinstance(self.schedule, int):
            return self.schedule
        else:
            return self.schedule[time % len(self.schedule)]

    def schedule_within(self, time: Union[int, List[int], Tuple[int, int]]) -> int:
        if isinstance(time, int):
            return self.schedule_at(time)
        if isinstance(time, tuple):
            times = list(range(time[0], time[1] + 1))
        else:
            times = time
        if isinstance(self.schedule, int):
            return self.schedule * len(times)
        return sum(self.schedule_at(t) for t in times)

    def set_schedule_at(self, time: int, value: int, n_steps: int) -> None:
        if isinstance(self.schedule, int):
            self.schedule = [self.schedule] * n_steps
        elif len(self.schedule) < n_steps:
            self.schedule = list(itertools.chain(*([self.schedule] * int(math.ceil(n_steps / len(self.schedule))))))
        self.schedule[time % len(self.schedule)] = value


class Consumer(SCMLAgent, ABC):
    """Base class of all consumer classes"""
    pass


class ScheduleDrivenConsumer(Consumer):
    """Consumer class"""

    def on_new_report(self, report: FinancialReport):
        pass

    def on_neg_request_rejected(self, req_id: str, by: Optional[List[str]]):
        pass

    def on_neg_request_accepted(self, req_id: str, mechanism: AgentMechanismInterface):
        pass

    def on_negotiation_failure(self, partners: List[str], annotation: Dict[str, Any],
                               mechanism: AgentMechanismInterface, state: MechanismState) -> None:
        pass

    def on_negotiation_success(self, contract: Contract, mechanism: AgentMechanismInterface) -> None:
        pass

    def on_contract_cancelled(self, contract: Contract, rejectors: List[str]) -> None:
        pass

    def on_contract_nullified(self, contract: Contract, bankrupt_partner: str, compensation: float) -> None:
        pass

    def on_agent_bankrupt(self, agent_id: str) -> None:
        pass

    def confirm_partial_execution(self, contract: Contract, breaches: List[Breach]) -> bool:
        return True

    def on_remove_cfp(self, cfp: 'CFP'):
        pass

    MAX_UNIT_PRICE = 1e5
    RELATIVE_MAX_PRICE = 2

    def __init__(self, profiles: Dict[int, ConsumptionProfile] = None
                 , negotiator_type='negmas.sao.AspirationNegotiator'
                 , consumption_horizon: Optional[int] = 20
                 , immediate_cfp_update: bool = True
                 , name=None):
        super().__init__(name=name)
        self.negotiator_type = get_class(negotiator_type, scope=globals())
        self.profiles: Dict[int, ConsumptionProfile] = dict()
        self.secured_quantities: Dict[int, int] = dict()
        if profiles is not None:
            self.set_profiles(profiles=profiles)
        self.consumption_horizon = consumption_horizon
        self.immediate_cfp_update = immediate_cfp_update

    def on_new_cfp(self, cfp: 'CFP') -> None:
        pass  # consumers never respond to CFPs

    def init(self):
        if self.consumption_horizon is None:
            self.consumption_horizon = self.awi.n_steps
        self.awi.register_interest(list(self.profiles.keys()))

    def set_profiles(self, profiles: Dict[int, ConsumptionProfile]):
        self.profiles = profiles if profiles is not None else dict()
        self.secured_quantities = dict(zip(profiles.keys(), itertools.repeat(0)))

    def register_product_cfps(self, p: int, t: int, profile: ConsumptionProfile):
        current_schedule = profile.schedule_at(t)
        product = self.products[p]
        awi: SCMLAWI = self.awi
        if current_schedule <= 0:
            awi.bb_remove(section='cfps', query={'publisher': self.id, 'time': t, 'product_index': p})
            return
        max_price = ScheduleDrivenConsumer.RELATIVE_MAX_PRICE * product.catalog_price if product.catalog_price is not None \
            else ScheduleDrivenConsumer.MAX_UNIT_PRICE
        cfps = awi.bb_query(section='cfps', query={'publisher': self.id, 'time': t, 'product': p})
        if cfps is not None and len(cfps) > 0:
            for _, cfp in cfps.items():
                if cfp.max_quantity != current_schedule:
                    cfp = CFP(is_buy=True, publisher=self.id, product=p
                              , time=t, unit_price=(0, max_price), quantity=(1, current_schedule))
                    awi.bb_remove(section='cfps', query={'publisher': self.id, 'time': t
                        , 'product': p})
                    awi.register_cfp(cfp)
                    break
        else:
            cfp = CFP(is_buy=True, publisher=self.id, product=p
                      , time=t, unit_price=(0, max_price), quantity=(1, current_schedule))
            awi.register_cfp(cfp)

    def step(self):
        if self.consumption_horizon is None:
            horizon = self.awi.n_steps
        else:
            horizon = min(self.awi.current_step + self.consumption_horizon + 1, self.awi.n_steps)
        for p, profile in self.profiles.items():
            for t in range(self.awi.current_step, horizon):  # + self.transportation_delay
                self.register_product_cfps(p=p, t=t, profile=profile)

    def confirm_contract_execution(self, contract: Contract) -> bool:
        return True

    @staticmethod
    def _qufun(outcome: Dict[str, Any], tau: float, profile: ConsumptionProfile):
        """The ufun value for quantity"""
        q, t = outcome['quantity'], outcome['time']
        y = profile.schedule_within(t)
        o = profile.overconsumption
        u = profile.underconsumption
        if q == 0 and y != 0:
            return 0.0
        if y <= 0:
            result = - o * ((q - y) ** tau)
        elif q > y:
            result = - o * (((q - y) / y) ** tau)
        elif q < y:
            result = - u * (((y - q) / y) ** tau)
        else:
            result = 1.0
        result = math.exp(result)
        if isinstance(result, complex):
            result = result.real
        if result is None:
            result = -1000.0
        return result

    def respond_to_negotiation_request(self, cfp: "CFP", partner: str) -> Optional[Negotiator]:
        if self.awi.is_bankrupt(partner):
            return None
        profile = self.profiles[cfp.product]
        if profile.cv == 0:
            alpha_u, alpha_q = profile.alpha_u, profile.alpha_q
        else:
            alpha_u, alpha_q = tuple(dirichlet((profile.alpha_u, profile.alpha_q), size=1)[0])
        beta_u = pos_gauss(profile.beta_u, profile.cv)
        tau_u = pos_gauss(profile.tau_u, profile.cv)
        tau_q = pos_gauss(profile.tau_q, profile.cv)
        ufun = ComplexWeightedUtilityFunction(ufuns=[
            MappingUtilityFunction(mapping=lambda x: 1 - x['unit_price'] ** tau_u / beta_u)
            , MappingUtilityFunction(mapping=functools.partial(ScheduleDrivenConsumer._qufun
                                                               , tau=tau_q, profile=profile))]
            , weights=[alpha_u, alpha_q], name=self.name + '_' + partner)
        ufun.reserved_value = -1500
        #ufun = normalize(, outcomes=cfp.outcomes, infeasible_cutoff=-1500)
        negotiator = self.negotiator_type(name=self.name + '*' + partner)
        negotiator.name = self.name + '_' + partner
        negotiator.utility_function = ufun
        return negotiator

    def set_renegotiation_agenda(self, contract: Contract, breaches: List[Breach]) -> Optional[RenegotiationRequest]:
        """
        Received by partners in ascending order of their total breach levels in order to set the
        renegotiation agenda when contract execution fails

        Args:

            contract: The contract that was breached about which re-negotiation is offered
            breaches: The list of breaches by all parties for the breached contract.

        Returns:

            None if renegotiation is not to be started, otherwise a re-negotiation agenda.

        """
        return None

    def respond_to_renegotiation_request(self, contract: Contract, breaches: List[Breach]
                                         , agenda: RenegotiationRequest) -> Optional[Negotiator]:
        """
        Called to respond to a renegotiation request

        Args:

            agenda: Renegotiatio agenda (issues to renegotiate about).
            contract: The contract that was breached
            breaches: All breaches on that contract

        Returns:

            None to refuse to enter the negotiation, otherwise, a negotiator to use for this negotiation.

        """
        return None

    def confirm_loan(self, loan: Loan, bankrupt_if_rejected: bool) -> bool:
        """called by the world manager to confirm a loan if needed by the buyer of a contract that is about to be
        breached"""
        return bankrupt_if_rejected

    def sign_contract(self, contract: Contract) -> Optional[str]:
        if contract is None:
            return None
        cfp: CFP = contract.annotation['cfp']
        agreement = contract.agreement  # type: ignore
        schedule = self.profiles[cfp.product].schedule_at(agreement['time'])
        if schedule - agreement['quantity'] < 0:
            return None
        return self.id

    def on_contract_signed(self, contract: Contract):
        if contract is None:
            return
        cfp: CFP = contract.annotation['cfp']
        agreement = contract.agreement  # type: ignore
        self.secured_quantities[cfp.product] += agreement['quantity']
        old_quantity = self.profiles[cfp.product].schedule_at(agreement['time'])
        new_quantity = old_quantity - agreement['quantity']
        t = agreement['time']
        self.profiles[cfp.product].set_schedule_at(time=t
                                                   , value=new_quantity
                                                   , n_steps=self.awi.n_steps)
        if self.immediate_cfp_update and new_quantity != old_quantity:
            self.register_product_cfps(p=cfp.product, t=t, profile=self.profiles[cfp.product])
        for negotiation in self._running_negotiations.values():
            self.notify(negotiation.negotiator, Notification(type='ufun_modified', data=None))






"""Common data-structures and classes used by all other modules.

This module does not import anything from the library except during type checking
"""
import typing
import uuid
from copy import deepcopy
from dataclasses import dataclass, field, fields
from typing import List, Optional, Any, TYPE_CHECKING

from .helpers import unique_name
from .java import to_java, to_dict, PYTHON_CLASS_IDENTIFIER

if TYPE_CHECKING:
    from .mechanisms import Mechanism
    from .outcomes import Issue, Outcome

__all__ = [
    'NamedObject',
    'AgentMechanismInterface',
    'MechanismState',
    'register_all_mechanisms',
    'NegotiatorInfo',
    '_ShadowAgentMechanismInterface',
]

_running_negotiations: typing.Dict[str, 'Mechanism'] = {}


@dataclass
class NegotiatorInfo:
    name: str
    """Name of this negotiator"""
    id: str
    """ID unique to this negotiator"""
    type: str
    """Type of the negotiator as a string"""


@dataclass
class MechanismState:
    """Encapsulates the mechanism state at any point"""
    running: bool = False
    """Whether the negotiation has started and did not yet finish"""
    started: bool = False
    """Whether the negotiation has started"""
    step: int = 0
    """The current round of the negotiation"""
    time: float = 0.0
    """The current real time of the negotiation."""
    relative_time: float = 0.0
    """A number in the period [0, 1] giving the relative time of the negotiation.
    Relative time is calculated as ``max(step/n_steps, time/time_limit)``.
    """
    broken: bool = False
    """True if the negotiation has started and ended with an END_NEGOTIATION"""
    timedout: bool = False
    """True if the negotiation was timedout"""
    agreement: Optional['Outcome'] = None
    """Agreement at the end of the negotiation (it is always None until an agreement is reached)"""
    n_negotiators: int = 0
    """Number of agents currently in the negotiation. Notice that this may change over time if the mechanism supports 
    dynamic entry"""
    has_error: bool = False
    """Does the mechanism have any errors"""
    error_details: str = ''
    """Details of the error if any"""

    @property
    def ended(self):
        return self.started and (self.broken or self.timedout or (self.agreement is not None))

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return self.__dict__.__repr__()

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __copy__(self):
        return AgentMechanismInterface(**self.__dict__)

    def __deepcopy__(self, memodict={}):
        d = {k: deepcopy(v) for k, v in self.__dict__.items()}
        return AgentMechanismInterface(**d)

    def __getitem__(self, item):
        """Makes the outcome type behave like a dict"""
        return self.__dict__[item]

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def asdict(self):
        """Converts the outcome to a dict containing all fields"""
        return {_.name: self.__dict__[_.name] for _ in fields(self)}

    class Java:
        implements = ['jnegmas.common.MechanismState']


@dataclass
class AgentMechanismInterface:
    """All information of a negotiation visible to negotiators."""
    id: str
    """Mechanism session ID. That is unique for all mechanisms"""
    n_outcomes: Optional[int]
    """Number of outcomes which may be None indicating infinity"""
    issues: List['Issue']
    """Negotiation issues as a list of `Issue` objects"""
    outcomes: Optional[List['Outcome']]
    """A lit of *all possible* outcomes for a negotiation. None if the number of outcomes is uncountable"""
    time_limit: float
    """The time limit in seconds for this negotiation session. None indicates infinity"""
    step_time_limit: float
    """The time limit in seconds for each step of this negotiation session. None indicates infinity"""
    n_steps: int
    """The allowed number of steps for this negotiation. None indicates infinity"""
    dynamic_entry: bool
    """Whether it is allowed for agents to enter/leave the negotiation after it starts"""
    max_n_agents: int
    """Maximum allowed number of agents in the session. None indicates no limit"""
    annotation: typing.Dict[str, Any] = field(default_factory=dict)
    """An arbitrary annotation as a `Dict[str, Any]` that is always available for all agents"""

    def random_outcomes(self, n: int = 1, astype: typing.Type['Outcome'] = dict) -> List['Outcome']:
        """
        A set of random outcomes from the issues of this negotiation

        Args:
            n: number of outcomes requested
            astype: A type to cast the resulting outcomes to.

        Returns:

            List[Outcome]: List of `n` or less outcomes

        """
        return _running_negotiations[self.id].random_outcomes(n=n, astype=astype)

    def discrete_outcomes(self, n_max: int = None, astype: typing.Type['Outcome'] = dict) -> List['Outcome']:
        """
        A discrete set of outcomes that spans the outcome space

        Args:
            n_max: The maximum number of outcomes to return. If None, all outcomes will be returned for discrete issues
            astype: A type to cast the resulting outcomes to.

        Returns:

            List[Outcome]: List of `n` or less outcomes

        """
        return _running_negotiations[self.id].discrete_outcomes(n_max=n_max, astype=astype)

    def outcome_index(self, outcome: 'Outcome') -> Optional[int]:
        """
        The index of an outcome

        Args:
            outcome: The outcome asked about

        Returns:

            int: The index of this outcome in the list of outcomes. Only valid if n_outcomes is finite and not None.
        """
        return _running_negotiations[self.id].outcome_index(outcome)

    @property
    def participants(self) -> List[NegotiatorInfo]:
        return _running_negotiations[self.id].participants

    @property
    def state(self) -> MechanismState:
        """
        Access the current state of the mechanism.

        Remarks:

            - Whenever a method receives a `AgentMechanismInterface` object, it can always access the *current* state of the
              protocol by accessing this property.

        """
        return _running_negotiations[self.id].state

    @property
    def requirements(self) -> dict:
        """
        The protocol requirements

        Returns:
            - A dict of str/Any pairs giving the requirements
        """
        return _running_negotiations[self.id].requirements

    @property
    def n_negotiators(self) -> int:
        """Syntactic sugar for state.n_agents"""
        return self.state.n_negotiators

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return self.__dict__.__repr__()

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        d1 = self.__dict__
        d2 = other.__dict__
        for k in d1.keys():
            if d2[k] != d1[k]:
                return False
        return True

    def __copy__(self):
        return AgentMechanismInterface(**self.__dict__)

    def __deepcopy__(self, memodict={}):
        d = {k: deepcopy(v) for k, v in self.__dict__.items()}
        return AgentMechanismInterface(**d)

    def __getitem__(self, item):
        """Makes the outcome type behave like a dict"""
        return self.__dict__[item]

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def asdict(self):
        """Converts the object to a dict containing all fields"""
        return {_.name: self.__dict__[_.name] for _ in fields(self)}


class _ShadowAgentMechanismInterface:
    """Used to represent an AMI to Java.
    """

    def randomOutcomes(self, n: int):
        return to_java(self.shadow.random_outcomes(n))

    def discreteOutcomes(self, nMax: int):
        return to_java(self.shadow.discrete_outcomes(n_max=nMax))

    def outcomeIndex(self, outcome) -> int:
        return to_java(self.shadow.outcome_index(outcome))

    def getParticipants(self) -> List[NegotiatorInfo]:
        return to_java(to_java(self.shadow.participants))

    def getOutcomes(self):
        return to_java(to_java(self.shadow.outcomes))

    def getState(self) -> MechanismState:
        return to_java(self.shadow.state)

    def getRequirements(self) -> typing.Dict[str, Any]:
        return to_java(self.shadow.requirements)

    def getNNegotiators(self) -> int:
        return self.shadow.n_negotiators

    def __init__(self, ami: AgentMechanismInterface):
        self.shadow = ami

    def to_java(self):
        return to_dict(self.shadow)
        
    class Java:
        implements = ['jnegmas.common.AgentMechanismInterface']


def register_all_mechanisms(mechanisms: typing.Dict[str, 'Mechanism']) -> None:
    """registers the running mechanisms. Used internally. **DO NOT CALL THIS.**"""
    global _running_negotiations
    _running_negotiations = mechanisms


class NamedObject(object):
    """The base class of all named entities.

    All named entities need to call this class's __init__() somewhere during initialization.

    Args:
        name (str): The given name of the entity. Notice that the class will add this to a  base that depends
                    on the child's class name.

    """

    def __init__(self, name: str = None) -> None:
        if name is not None:
            name = str(name)
        self.__uuid = (f'{name}-' if name is not None else "") + str(uuid.uuid4())
        if name is None or len(name) == 0:
            name = unique_name('', add_time=False, rand_digits=16)
        self.__name = name
        super().__init__()

    @classmethod
    def create(cls, *args, **kwargs):
        """Creates an object and returns a proxy to it."""
        return cls(*args, **kwargs)

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name):
        self.__name = name

    @property
    def uuid(self):
        return self.__uuid

    @uuid.setter
    def uuid(self, uuid):
        self.__uuid = uuid

    @property
    def id(self):
        return self.__uuid

    @id.setter
    def id(self, id):
        self.__uuid = id



#!/usr/bin/env python
"""The NegMAS universal command line tool"""
import json
import os
import traceback
from functools import partial
from math import factorial
from pathlib import Path
from pprint import pformat
from time import perf_counter
import urllib.request

import click
import pandas as pd
import progressbar
import yaml
from tabulate import tabulate

import negmas
from negmas import save_stats
from negmas.apps.scml import *
from negmas.helpers import humanize_time, unique_name
from negmas.java import init_jnegmas_bridge, jnegmas_bridge_is_running

try:
    # disable a warning in yaml 1b1 version
    yaml.warnings({'YAMLLoadWarning': False})
except:
    pass

n_completed = 0
n_total = 0

JNEGMAS_JAR_NAME = 'jnegmas-0.2.1-all.jar'
GENIUS_JAR_NAME = 'genius-8.0.4-bridge.jar'


def print_progress(_, i, n) -> None:
    """Prints the progress of a tournament"""
    global n_completed, n_total
    n_completed = i + 1
    n_total = n
    print(f'{n_completed:04} of {n:04} worlds completed ({n_completed / n:0.2%})', flush=True)


def print_world_progress(world) -> None:
    """Prints the progress of a world"""
    step = world.current_step + 1
    s = f'World# {n_completed:04}: {step:04}  of {world.n_steps:04} ' \
        f'steps completed ({step / world.n_steps:0.2f}) '
    if n_total > 0:
        s += f'TOTAL: ({n_completed + step / world.n_steps / n_total:0.2f})'
    print(s, flush=True)


click.option = partial(click.option, show_default=True)


@click.group()
def cli():
    pass


@cli.command(help='Run a tournament between candidate agent types')
@click.option('--name', '-n', default='random',
              help='The name of the tournament. The special value "random" will result in a random name')
@click.option('--steps', '-s', default=60, help='Number of steps.')
@click.option('--ttype', '--tournament-type', '--tournament', default='anac2019collusion'
    , help='The config to use. Default is ANAC 2019. Options supprted are anac2019std, anac2019collusion, '
           'anac2019sabotage')
@click.option('--timeout', '-t', default=0, help='Timeout after the given number of seconds (0 for infinite)')
@click.option('--configs', default=5, help='Number of unique configurations to generate.')
@click.option('--runs', default=2, help='Number of runs for each configuration')
@click.option('--max-runs', default=-1, help='Maximum total number of runs. Zero or negative numbers mean no limit')
@click.option('--competitors'
    , default='negmas.apps.scml.DoNothingFactoryManager;negmas.apps.scml.GreedyFactoryManager'
    , help='A semicolon (;) separated list of agent types to use for the competition.')
@click.option('--jcompetitors', '--java-competitors'
    , default=''
    , help='A semicolon (;) separated list of agent types to use for the competition.')
@click.option('--parallel/--serial', default=True, help='Run a parallel/serial tournament on a single machine')
@click.option('--distributed/--single-machine', default=False, help='Run a distributed tournament using dask')
@click.option('--log', '-l', default='~/negmas/logs/tournaments',
              help='Default location to save logs (A folder will be created under it)')
@click.option('--verbosity', default=1, help='verbosity level (from 0 == silent to 1 == world progress)')
@click.option('--configs-only/--run', default=False, help='configs_only')
@click.option('--reveal-names/--hidden-names', default=True, help='Reveal agent names (should be used only for '
                                                                  'debugging)')
@click.option('--ip', default='127.0.0.1', help='The IP address for a dask scheduler to run the distributed tournament.'
                                                ' Effective only if --distributed')
@click.option('--port', default=8786, help='The IP port number a dask scheduler to run the distributed tournament.'
                                           ' Effective only if --distributed')
@click.option('--compact/--debug', default=True, help='If True, effort is exerted to reduce the memory footprint which'
                                                      'includes reducing logs dramatically.')
def tournament(name, steps, parallel, distributed, ttype, timeout, log, verbosity, configs_only,
               reveal_names, ip, port, runs, configs, max_runs, competitors, jcompetitors, compact):
    if timeout <= 0:
        timeout = None
    if name == 'random':
        name = None
    if max_runs <= 0:
        max_runs = None

    worlds_per_config = None if max_runs is None else max_runs / (configs * runs)

    parallelism = 'distributed' if distributed else 'parallel' if parallel else 'serial'
    all_competitors = competitors.split(';')
    all_competitors_params = [dict() for _ in range(len(all_competitors))]
    if jcompetitors is not None and len(jcompetitors) > 0:
        jcompetitor_params = [{'java_class_name': _} for _ in jcompetitors.split(';')]
        jcompetitors = ['negmas.apps.scml.JavaFactoryManager'] * len(jcompetitor_params)
        all_competitors += jcompetitors
        all_competitors_params += jcompetitor_params
        print('You are using some Java agents. The tournament MUST run serially')
        parallelism = 'serial'
        if not jnegmas_bridge_is_running():
            print('Error: You are using java competitors but jnegmas bridge is not running\n\nTo correct this issue'
                  ' run the following command IN A DIFFERENT TERMINAL because it will block:\n\n'
                  '$ negmas jnegmas')
            exit(1)
    prog_callback = print_world_progress if verbosity > 1 and not distributed else None

    recommended = runs * configs * factorial(len(all_competitors) * (1 if "std" in ttype else 3))
    if worlds_per_config < 1:
        print(f'You need at least {(configs * runs)} runs even with a single permutation of managers.'
              f'.\n\nSet --max-runs to at least {(configs * runs)} (Recommended {recommended})')
        return
    worlds_per_config = worlds_per_config if worlds_per_config is None else int(worlds_per_config)
    if max_runs < recommended:
        print(f'You are running {max_runs} worlds only but it is recommended to set {max_runs} to at least '
              f'{recommended}. Will continue')

    start = perf_counter()
    if ttype.lower() == 'anac2019std':
        results = anac2019_std(competitors=all_competitors
                               , competitor_params=all_competitors_params
                               , agent_names_reveal_type=reveal_names
                               , tournament_path=log, total_timeout=timeout
                               , parallelism=parallelism, scheduler_ip=ip, scheduler_port=port
                               , world_progress_callback=prog_callback
                               , name=name, verbose=verbosity > 0, n_runs_per_world=runs, n_configs=configs
                               , max_worlds_per_config=worlds_per_config
                               , configs_only=configs_only
                               , n_steps=steps, compact=compact)
    elif ttype.lower() in ('anac2019collusion', 'anac2019'):
        results = anac2019_collusion(competitors=all_competitors
                                     , competitor_params=all_competitors_params
                                     , agent_names_reveal_type=reveal_names
                                     , tournament_path=log, total_timeout=timeout
                                     , parallelism=parallelism, scheduler_ip=ip, scheduler_port=port
                                     , world_progress_callback=prog_callback
                                     , name=name, verbose=verbosity > 0, n_runs_per_world=runs, n_configs=configs
                                     , max_worlds_per_config=worlds_per_config
                                     , configs_only=configs_only
                                     , n_steps=steps, compact=compact)
    else:
        print(f'{ttype.lower()} tournament type is not supported')
        exit(1)
    if configs_only:
        print(f'Saved all configs to {str(results)}')
        print(f'Finished in {humanize_time(perf_counter() - start)} [config-only]')
        return
    print(tabulate(results.total_scores, headers='keys', tablefmt='psql'))
    print(f'Finished in {humanize_time(perf_counter() - start)} [{parallelism}]')


@cli.command(help='Run an SCML world simulation')
@click.option('--steps', default=120, help='Number of steps.')
@click.option('--levels', default=3, help='Number of intermediate production levels (processes). '
                                          '-1 means a single product and no factories.')
@click.option('--neg-speedup', default=21, help='Negotiation Speedup.')
@click.option('--negotiator', default='negmas.sao.AspirationNegotiator',
              help='Negotiator type to use for builtin agents.')
@click.option('--min-consumption', default=3, help='The minimum number of units consumed by each consumer at every '
                                                   'time-step.')
@click.option('--max-consumption', default=5, help='The maximum number of units consumed by each consumer at every '
                                                   'time-step.')
@click.option('--agents', default=5, help='Number of agents (miners/negmas.consumers) per production level')
@click.option('--horizon', default=20, help='Consumption horizon.')
@click.option('--transport', default=0, help='Transportation Delay.')
@click.option('--time', default=60 * 90, help='Total time limit.')
@click.option('--neg-time', default=60 * 4, help='Time limit per single negotiation')
@click.option('--neg-steps', default=20, help='Number of rounds per single negotiation')
@click.option('--sign', default=1, help='The default delay between contract conclusion and signing')
@click.option('--guaranteed', default=False, help='Whether to only sign contracts that are guaranteed not to cause '
                                                  'breaches')
@click.option('--lines', default=10, help='The number of lines per factory')
@click.option('--retrials', default=5, help='The number of times an agent re-tries on failed negotiations')
@click.option('--use-consumer', default=True, help='Use internal consumer object in factory managers')
@click.option('--max-insurance', default=100, help='Use insurance against partner in factory managers up to this '
                                                   'premium')
@click.option('--riskiness', default=0.0, help='How risky is the default factory manager')
@click.option('--log', default='~/negmas/logs',
              help='Default location to save logs (A folder will be created under it)')
@click.option('--compact/--debug', default=True, help='If True, effort is exerted to reduce the memory footprint which'
                                                      'includes reducing logs dramatically.')
def scml(steps, levels, neg_speedup, negotiator, agents, horizon, min_consumption, max_consumption
         , transport, time, neg_time
         , neg_steps, sign, guaranteed, lines, retrials, use_consumer, max_insurance, riskiness, log, compact):
    params = {
        "steps": steps,
        "levels": levels,
        "neg_speedup": neg_speedup,
        "negotiator": negotiator,
        "agents": agents,
        "horizon": horizon,
        "min_consumption": min_consumption,
        "max_consumption": max_consumption,
        "transport": transport,
        "time": time,
        "neg_time": neg_time,
        "neg_steps": neg_steps,
        "sign": sign,
        "guaranteed": guaranteed,
        "lines": lines,
        "retrials": retrials,
        "use_consumer": use_consumer,
        "max_insurance": max_insurance,
        "riskiness": riskiness
    }
    neg_speedup = neg_speedup if neg_speedup is not None and neg_speedup > 0 else None
    if min_consumption == max_consumption:
        consumption = min_consumption
    else:
        consumption = (min_consumption, max_consumption)
    customer_kwargs = {'negotiator_type': negotiator, 'consumption_horizon': horizon}
    miner_kwargs = {'negotiator_type': negotiator, 'n_retrials': retrials}
    factory_kwargs = {'negotiator_type': negotiator, 'n_retrials': retrials
        , 'sign_only_guaranteed_contracts': guaranteed, 'use_consumer': use_consumer
        , 'riskiness': riskiness, 'max_insurance_premium': max_insurance}
    if log.startswith('~/'):
        log_dir = Path.home() / log[2:]
    else:
        log_dir = Path(log)
    log_dir = log_dir / unique_name(base='scml', add_time=True, rand_digits=0)
    log_dir = log_dir.absolute()
    os.makedirs(log_dir, exist_ok=True)
    log_file_name = str(log_dir / 'log.txt')
    exception = None
    world = SCMLWorld.chain_world(log_file_name=log_file_name, n_steps=steps
                                  , negotiation_speed=neg_speedup
                                  , n_intermediate_levels=levels
                                  , n_miners=agents
                                  , n_consumers=agents
                                  , n_factories_per_level=agents
                                  , consumption=consumption
                                  , consumer_kwargs=customer_kwargs
                                  , miner_kwargs=miner_kwargs
                                  , default_manager_params=factory_kwargs
                                  , transportation_delay=transport, time_limit=time, neg_time_limit=neg_time
                                  , neg_n_steps=neg_steps, default_signing_delay=sign
                                  , n_lines_per_factory=lines, compact=compact)
    failed = False
    strt = perf_counter()
    try:
        for i in progressbar.progressbar(range(world.n_steps), max_value=world.n_steps):
            elapsed = perf_counter() - strt
            if world.time_limit is not None and elapsed >= world.time_limit:
                break
            if not world.step():
                break
    except Exception:
        exception = traceback.format_exc()
        failed = True
    elapsed = perf_counter() - strt

    def print_and_log(s):
        world.logdebug(s)
        print(s)

    world.logdebug(f'{pformat(world.stats, compact=True)}')
    world.logdebug(f'=================================================\n'
                   f'steps: {steps}, horizon: {horizon}, time: {time}, levels: {levels}, agents_per_level: '
                   f'{agents}, lines: {lines}, guaranteed: {guaranteed}, negotiator: {negotiator}\n'
                   f'consumption: {consumption}'
                   f', transport_to: {transport}, sign: {sign}, speedup: {neg_speedup}, neg_steps: {neg_steps}'
                   f', retrials: {retrials}'
                   f', neg_time: {neg_time}\n'
                   f'==================================================')

    save_stats(world=world, log_dir=log_dir, params=params)

    if len(world.saved_contracts) > 0:
        data = pd.DataFrame(world.saved_contracts)
        data = data.sort_values(['delivery_time'])
        data = data.loc[data.signed_at >= 0, ['seller_type', 'buyer_type', 'seller_name', 'buyer_name', 'delivery_time'
                        , 'unit_price', 'quantity', 'product_name', 'n_neg_steps', 'signed_at']]
        data.columns = ['seller_type', 'buyer_type', 'seller', 'buyer', 't', 'price', 'q', 'product', 'steps'
                        , 'signed']
        print_and_log(tabulate(data, headers='keys', tablefmt='psql'))
        n_executed = sum(world.stats['n_contracts_executed'])
        n_negs = sum(world.stats["n_negotiations"])
        n_contracts = len(world.saved_contracts)
        winners = [f'{_.name} gaining {world.a2f[_.id].balance / world.a2f[_.id].initial_balance - 1.0:0.0%}'
                   for _ in world.winners]
        print_and_log(f'{n_contracts} contracts :-) [N. Negotiations: {n_negs}, Agreement Rate: '
                      f'{world.agreement_rate:0.0%}]')
        print_and_log(f'Cancelled: {world.cancellation_rate:0.0%}, Executed: {world.contract_execution_fraction:0.0%}'
                      f', Breached: {world.breach_rate:0.0%}, N. Executed: {n_executed}, Business size: '
                      f'{world.business_size}\n'
                      f'Winners: {winners}\n'
                      f'Running Time {humanize_time(elapsed)}')
    else:
        print_and_log('No contracts! :-(')
        print_and_log(f'Running Time {humanize_time(elapsed)}')

    if failed:
        print(exception)
        world.logdebug(exception)
        print(f'FAILED at step {world.current_step} of {world.n_steps}\n')


@cli.command(help='Start the bridge to genius (to use GeniusNegotiator)')
@click.option('--path', '-p', default='auto', help='Path to genius-8.0.4.jar with embedded NegLoader. Use "auto" to '
                                                   'read the path from ~/negmas/config.json'
                                                   '\n\tConfig key is genius_bridge_jar'
                                                   '\nYou can download this jar from: '
                                                   'http://www.yasserm.com/scml/genius-8.0.4-bridge.jar')
@click.option('--port', '-r', default=0, help='Port to run the NegLoader on. Pass 0 for the default value')
@click.option('--force/--no-force', default=False, help='Force trial even if an earlier instance exists')
def genius(path, port, force):
    negmas.init_genius_bridge(path=path if path != 'auto' else None, port=port, force=force)


@cli.command(help='Start the bridge to JNegMAS (to use Java agents in worlds)')
@click.option('--path', '-p', default='auto', help='Path to jnegmas*.jar with. Use "auto" to '
                                                   'read the path from ~/negmas/config.json.'
                                                   '\n\tConfig key is jnegmas_jar'
                                                   '\nYou can download the latest version of this jar from: '
                                                   'http://www.yasserm.com/scml/jnegmas-all.jar')
@click.option('--port', '-r', default=0, help='Port to run the jnegmas on. Pass 0 for the default value')
def jnegmas(path, port):
    init_jnegmas_bridge(path=path if path != 'auto' else None, port=port)
    input('Press C^c to quit. You may also need to kill any remaining java processes manually')


def download_and_set(key, url, file_name):
    """
    Downloads a file and sets the corresponding key in ~/negmas/config.json

    Args:
        key: Key name in config.json
        url: URL to download from
        file_name: file name

    Returns:

    """
    config_path = Path.home() / 'negmas'
    jar_path = config_path / 'files'
    jar_path.mkdir(parents=True, exist_ok=True)
    jar_path = jar_path / file_name
    urllib.request.urlretrieve(url, jar_path)
    config = {}
    config_file = config_path / 'config.json'
    if config_file.exists():
        with open(config_file, 'r') as f:
            config = json.load(f)
    config[key] = str(jar_path)
    with open(config_file, 'w') as f:
        json.dump(config, fp=f, sort_keys=True, indent=4)


@cli.command(help='Downloads jnegmas and updates your settings')
def jnegmas_setup():
    url = f'http://www.yasserm.com/scml/{JNEGMAS_JAR_NAME}'
    print(f'Downloading: {url}', end='', flush=True)
    download_and_set(key='jnegmas_jar', url=url, file_name='jnegmas.jar')
    print(' done successfully')


@cli.command(help='Downloads the genius bridge and updates your settings')
def genius_setup():
    url = f'http://www.yasserm.com/scml/{GENIUS_JAR_NAME}'
    print(f'Downloading: {url}', end='', flush=True)
    download_and_set(key='genius_bridge_jar', url=url, file_name='jnegmas.jar')
    print(' done successfully')


if __name__ == '__main__':
    cli()



import copy
from pprint import pprint
from typing import List, Dict

import numpy as np
import pkg_resources
import pytest

from negmas.apps.scml import *
from negmas.apps.scml import InputOutput, Job, FactoryStatusUpdate, GreedyScheduler, ProductionFailure
from negmas.situated import Contract


# def test_can_create_a_random_scml_world():
#     world = SCMLWorld.random()
#     assert len(world.products) > 0
#     assert len(world.processes) > 0
#     assert len(world.stats['market_size']) == 0
#     winners = [_.name for _ in world.winners]
#     assert len(winners) == len(world.factory_managers)
#     for x in ('n_contract_executions', 'n_breaches', 'n_cfps_on_board', 'n_cfps'):
#         assert len(world.stats[x]) == 0


# def test_can_run_a_random_small_scml_world():
#     world = SCMLWorld.random_small(log_file_name='', n_steps=200)
#     world.run()
#     print('')
#     for key in sorted(world.stats.keys()):
#         print(f'{key}:{world.stats[key]}')
#     winners = [_.name for _ in world.winners]
#     print('Winners: ', winners)
#     print('Losers: ', [_.name for _ in world.factory_managers if _.name not in winners])
#     data = pd.DataFrame(data=world.saved_contracts).loc[:, ["buyer", "seller", "product", "quantity", "delivery_time"
#     , "unit_price", "penalty", "signing_delay", "concluded_at", "signed_at", "issues", "cfp"]]
#     data.to_csv(f'{logdir()}/contracts.csv')
#
#
# def test_can_run_a_random_small_scml_world_with_delayed_signing():
#     world = SCMLWorld.random_small(log_file_name='', n_steps=20, default_signing_delay=1)
#     world.run()
#     print('')
#     for key in sorted(world.stats.keys()):
#         print(f'{key}:{world.stats[key]}')
#     winners = [_.name for _ in world.winners]
#     print('Winners: ', winners)
#     print('Losers: ', [_.name for _ in world.factory_managers if _.name not in winners])
#     data = pd.DataFrame(data=world.saved_contracts).loc[:, ["buyer", "seller", "product", "quantity", "delivery_time"
#     , "unit_price", "penalty", "signing_delay", "concluded_at", "signed_at", "issues", "cfp"]]
#     data.to_csv(f'{logdir()}/contracts.csv')
#
#
# def test_can_run_a_random_small_scml_world_with_money_resolution():
#     world = SCMLWorld.random_small(log_file_name='', n_steps=20, money_resolution=1)
#     world.run()
#     print('')
#     for key in sorted(world.stats.keys()):
#         print(f'{key}:{world.stats[key]}')
#     winners = [_.name for _ in world.winners]
#     print('Winners: ', winners)
#     print('Losers: ', [_.name for _ in world.factory_managers if _.name not in winners])
#     data = pd.DataFrame(data=world.saved_contracts).loc[:, ["buyer", "seller", "product", "quantity", "delivery_time"
#     , "unit_price", "penalty", "signing_delay", "concluded_at", "signed_at", "issues", "cfp"]]
#     data.to_csv(f'{logdir()}/contracts.csv')


def logdir():
    return pkg_resources.resource_filename('negmas', resource_name='tests')


def test_can_run_a_random_tiny_scml_world():
    world = SCMLWorld.chain_world(log_file_name='', n_steps=5, n_factories_per_level=1
                                  , consumer_kwargs={'negotiator_type': 'negmas.sao.NiceNegotiator', 'consumption_horizon': 2}
                                  # , factory_kwargs={'max_insurance_premium': 100}
                                  , miner_kwargs={'negotiator_type': 'negmas.sao.NiceNegotiator'}
                                  )
    world.run()
    # print('')
    # for key in sorted(world.stats.keys()):
    #     print(f'{key}:{world.stats[key]}')
    winners = [_.name for _ in world.winners]
    # print('Winners: ', winners)
    # print('Losers: ', [_.name for _ in world.factory_managers if _.name not in winners])
    # if len(world.saved_contracts) > 0:
    #     data = pd.DataFrame(data=world.saved_contracts).loc[:, ["buyer", "seller", "product", "quantity", "delivery_time"
    #     , "unit_price", "penalty", "signing_delay", "concluded_at", "signed_at", "issues", "cfp"]]
    #     data.to_csv(f'{logdir()}/contracts.csv')


def test_can_run_a_random_tiny_scml_world_no_immediate():
    world = SCMLWorld.chain_world(log_file_name='', n_steps=5)
    world.run()


def test_can_run_a_random_tiny_scml_world_with_insurance():
    world = SCMLWorld.chain_world(log_file_name='', n_steps=5
                                  # , factory_kwargs={'max_insurance_premium': 1e6}
                                  , consumer_kwargs={'negotiator_type': 'negmas.sao.NiceNegotiator'}
                                  , miner_kwargs={'negotiator_type': 'negmas.sao.NiceNegotiator'})
    world.run()

# @settings(max_examples=50)
# @given(horizon=st.sampled_from([None, 10, 50, 100, 500]),
#        immediate=st.booleans(),
#        signing_delay=st.integers(0, 3),
#        n_factory_levels=st.integers(1, 5),
#        n_factories_per_level=st.integers(1, 5)
#        )
# def test_can_run_a_random_tiny_scml_world_with_n_factories(horizon, immediate, signing_delay
#                                                            , n_factory_levels, n_factories_per_level):
#     n_steps = 500
#     world = SCMLWorld.tiny(n_intermediate_levels=n_factory_levels - 1, log_file_name='', n_steps=n_steps
#                            , n_factories_per_level=n_factories_per_level
#                            , default_signing_delay=signing_delay
#                            , consumer_kwargs={'immediate_cfp_update': immediate, 'consumption_horizon': horizon
#             , 'negotiator_type': 'NiceNegotiator'}
#                            , miner_kwargs={'negotiator_type': 'NiceNegotiator'
#             , 'use_immediate_negotiations': immediate}
#                            , factory_kwargs={'max_insurance_premium': -1
#             , 'use_immediate_negotiations': immediate}
#                            )
#     world.run()
#     print('')
#     for key in sorted(world.stats.keys()):
#         print(f'{key}:{world.stats[key]}')
#     winners = [_.name for _ in world.winners]
#     print('Winners: ', winners)
#     print('Losers: ', [_.name for _ in world.factory_managers if _.name not in winners])
#     # assert sum(world.stats['n_contracts_concluded']) > 0
#     # assert sum(world.stats['n_contracts_signed']) > 0
#     # assert len(winners) < len(world.factory_managers)


def test_can_run_a_random_tiny_scml_world_with_linear_production():
    horizon = None
    signing_delay = 0
    n_factory_levels = 0
    n_factories_per_level = 2
    n_steps = 10
    world = SCMLWorld.chain_world(n_intermediate_levels=n_factory_levels - 1, log_file_name='', n_steps=n_steps
                                  , n_factories_per_level=n_factories_per_level
                                  , default_signing_delay=signing_delay
                                  , consumer_kwargs={'consumption_horizon': horizon
                                              , 'negotiator_type': 'negmas.sao.NiceNegotiator'}
                                  , miner_kwargs={'negotiator_type': 'negmas.sao.NiceNegotiator'}
                                  )
    world.run()
    assert sum(world.stats['n_contracts_concluded']) > 0


def test_can_run_a_random_tiny_scml_world_with_no_factory():
    n_steps = 10
    world = SCMLWorld.chain_world(n_intermediate_levels=-1, log_file_name='', n_steps=n_steps
                                  , negotiation_speed=None)
    world.run()
    # print('')
    # for key in sorted(world.stats.keys()):
    #     print(f'{key}:{world.stats[key]}')
    assert world.stats['n_negotiations'][0] >= n_steps - 1, "at least n negotiations occur where n is the number of " \
                                                            "steps - 1"
    assert world.stats['n_negotiations'][2:] == [0] * (len(world.stats['n_negotiations']) - 2), "All negotiations " \
                                                                                                "happen in steps 0, 1"
    assert sum(world.stats['n_contracts_concluded']) >= sum(world.stats['n_contracts_signed']), "some contracts signed"
    assert sum(world.stats['n_breaches']) == 0, "No breaches"
    assert sum(world.stats['market_size']) == 0, "No change in the market size"

    # data = pd.DataFrame(data=world.saved_contracts).loc[:, ["buyer", "seller", "product", "quantity", "delivery_time"
    # , "unit_price", "penalty", "signing_delay", "concluded_at", "signed_at", "issues", "cfp"]]
    # data.to_csv(f'{logdir()}/contracts.csv')


def test_can_run_a_random_tiny_scml_world_with_no_factory_finite_horizon():
    n_steps = 5
    horizon = n_steps // 2
    world = SCMLWorld.chain_world(n_intermediate_levels=-1, log_file_name='', n_steps=n_steps
                                  , consumer_kwargs={'consumption_horizon': horizon})
    world.run()
    # print('')
    # for key in sorted(world.stats.keys()):
    #     print(f'{key}:{world.stats[key]}')
    #assert world.stats['n_negotiations'][0] >= horizon - 1, "at least n negotiations occur where n is the number of " \
    #                                                        "horizon - 1"
    assert sum(world.stats['n_contracts_concluded']) >= sum(world.stats['n_contracts_signed']), "all contracts signed"
    assert sum(world.stats['n_breaches']) == 0, "No breaches"
    assert sum(world.stats['market_size']) == 0, "No change in the market size"

    # data = pd.DataFrame(data=world.saved_contracts).loc[:, ["buyer", "seller", "product", "quantity", "delivery_time"
    # , "unit_price", "penalty", "signing_delay", "concluded_at", "signed_at", "issues", "cfp"]]
    # data.to_csv(f'{logdir()}/contracts.csv')


def test_can_run_a_random_tiny_scml_world_with_no_factory_with_delay():
    n_steps = 10
    world = SCMLWorld.chain_world(n_intermediate_levels=-1, log_file_name='', n_steps=n_steps, default_signing_delay=1)
    world.run()
    # print('')
    # for key in sorted(world.stats.keys()):
    #     print(f'{key}:{world.stats[key]}')
    assert world.stats['n_negotiations'][0] >= n_steps - 1, "at least n negotiations occur where n is the number of " \
                                                            "steps - 1"
    #assert world.stats['n_negotiations'][1:] == [0] * (len(world.stats['n_negotiations']) - 1), "All negotiations " \
    #                                                                                           "happen in step 0"
    assert sum(world.stats['n_contracts_concluded']) >= sum(world.stats['n_contracts_signed']), "all contracts signed"
    assert sum(world.stats['n_breaches']) == 0, "No breaches"
    assert sum(world.stats['market_size']) == 0, "No change in the market size"

    # data = pd.DataFrame(data=world.saved_contracts).loc[:, ["buyer", "seller", "product", "quantity", "delivery_time"
    # , "unit_price", "penalty", "signing_delay", "concluded_at", "signed_at", "issues", "cfp"]]
    # data.to_csv(f'{logdir()}/contracts.csv')


def test_can_run_a_random_tiny_scml_world_with_no_factory_with_delay_no_immediate_neg():
    n_steps = 10
    horizon = 4
    world = SCMLWorld.chain_world(n_intermediate_levels=-1, log_file_name='', n_steps=n_steps
                                  , default_signing_delay=1
                                  , consumer_kwargs={'immediate_cfp_update': False
                                                           , 'consumption_horizon': horizon})
    world.run()
    # print('')
    # for key in sorted(world.stats.keys()):
    #     print(f'{key}:{world.stats[key]}')
    #assert world.stats['n_negotiations'][0] >= horizon - 1, "at least n negotiations occur where n is the number of " \
    #                                                        "horizon - 1"
    assert sum(world.stats['n_contracts_concluded']) >= world.stats['n_contracts_concluded'][-1] + \
           sum(world.stats['n_contracts_signed']), "some contracts signed"
    assert sum(world.stats['n_breaches']) == 0, "No breaches"
    assert sum(world.stats['market_size']) == 0, "No change in the market size"


if __name__ == '__main__':
    pytest.main(args=[__file__])
